
		/*********************************************************
		** Uses some scan-line and point-fitting techniques to
		** calculate a circle and line clipping equations for the
		** valid/invalid data boundary in the fundus images.
		**
		** Coded by Adam Hoover in October '97.
		** Ported to windows in September '98.
		*********************************************************/

#include <stdio.h>
#include <math.h>
#include <windows.h>
#include "globals.h"

#define CROSSOVER_THRESHOLD	30



void OverlayStareMask(unsigned char *InputImage,
					  int ROWS,int COLS,int nbytes,
					  unsigned char *OutputImage)

{
int		r,c,b;
double	radius;

for (r=0; r<ROWS; r++)
  for (c=0; c<COLS; c++)
	for (b=0; b<nbytes; b++)
      {
      OutputImage[(r*COLS+c)*nbytes+b]=0; /* change to 0 for processing, 255 for making figures */
      if (r < MaskTopRow)
        continue;
      if (r > MaskBottomRow)
        continue;
      radius=sqrt(SQR((double)r-MaskCircle[1])+SQR((double)c-MaskCircle[0]));
      if (radius > MaskCircle[2])
        continue;
      OutputImage[(r*COLS+c)*nbytes+b]=InputImage[(r*COLS+c)*nbytes+b];
      }
}




        /*********************************************************
        ** Stores mask as three global variables:  MaskTopRow,
		** MaskBottomRow and MaskCircle.  The MaskCircle is a
		** three element array (center row, center column, radius).
        *********************************************************/

void ComputeStareMask(unsigned char *Raw,
					  int ROWS,int COLS,
					  unsigned char *MaskImage)	  // returns image if not NULL

{
int				x_coords[400],y_coords[400],total_points,r,c;
int				largest,largest_index;
int				*top_line_count,*bottom_line_count;
int				RedGrad,GreenGrad,BlueGrad;


if ((top_line_count=(int *)calloc(ROWS,sizeof(int))) == NULL  ||
	(bottom_line_count=(int *)calloc(ROWS,sizeof(int))) == NULL)
  {
  MessageBox(MainWnd,"Error -- out of memory","Crash",MB_OK | MB_APPLMODAL);
  return;
  }

        /*********************************************************
        ** Looking downward from top of image, and upward from
		** bottom of image, but only in the middle 100 pixels of
		** the image, find the crossover from invalid to valid
		** in each column.  The crossover is taken as the mode
		** of the samples, inward 3 rows in case there is a slant.
        *********************************************************/

for (r=0; r<ROWS; r++)
  top_line_count[r]=bottom_line_count[r]=0;
for (c=COLS/2-50; c<COLS/2+50; c++)
  {
  largest=largest_index=0;
  for (r=3; r<ROWS/4; r++)
    {
    BlueGrad=Raw[(r*COLS+c)*3+2]-Raw[((r-3)*COLS+c)*3+2];
    GreenGrad=Raw[(r*COLS+c)*3+1]-Raw[((r-3)*COLS+c)*3+1];
    RedGrad=Raw[(r*COLS+c)*3+0]-Raw[((r-3)*COLS+c)*3+0];
	if (RedGrad+GreenGrad+BlueGrad > largest)
	  {
	  largest=RedGrad+GreenGrad+BlueGrad;
	  largest_index=r;
	  }
    }
  if (largest > CROSSOVER_THRESHOLD)
	{
    top_line_count[largest_index]++;
	}
  largest=largest_index=0;
  for (r=ROWS-4; r>=ROWS*3/4; r--)
    {
    RedGrad=Raw[(r*COLS+c)*3+0]-Raw[((r+3)*COLS+c)*3+0];
    GreenGrad=Raw[(r*COLS+c)*3+1]-Raw[((r+3)*COLS+c)*3+1];
    BlueGrad=Raw[(r*COLS+c)*3+2]-Raw[((r+3)*COLS+c)*3+2];
	if (RedGrad+GreenGrad+BlueGrad > largest)
	  {
	  largest=RedGrad+GreenGrad+BlueGrad;
	  largest_index=r;
	  }
    }
  if (largest > CROSSOVER_THRESHOLD)
	{
    bottom_line_count[largest_index]++;
	}
  }
MaskTopRow=MaskBottomRow=0;
for (r=1; r<ROWS; r++)
  {
  if (top_line_count[r] > top_line_count[MaskTopRow])
    MaskTopRow=r;
  if (bottom_line_count[r] > bottom_line_count[MaskBottomRow])
    MaskBottomRow=r;
  }
MaskTopRow+=3;
MaskBottomRow-=3;


        /*********************************************************
        ** Looking rightward from left side of image, and leftward
		** from right side of image, but only in the middle 200
		** pixels of each side, find the crossover from invalid to
		** valid in each row.  These points are then fed into a
		** circle fitting routine (based upon least squares).
		** The radius of the circle is decreased by 2, to error
		** on the side of caution.
        *********************************************************/

total_points=0;
/* for (r=ROWS/2-100; r<ROWS/2+100; r++) */
for (r=100; r<ROWS-100; r++)
  {
  if (r == 150) r=ROWS-150;
  for (c=3; c<COLS/8; c++)
    {
    RedGrad=Raw[(r*COLS+c)*3+0]-Raw[(r*COLS+c-3)*3+0];
    GreenGrad=Raw[(r*COLS+c)*3+1]-Raw[(r*COLS+c-3)*3+1];
    BlueGrad=Raw[(r*COLS+c)*3+2]-Raw[(r*COLS+c-3)*3+2];
    if (RedGrad+GreenGrad+BlueGrad > CROSSOVER_THRESHOLD)
      break;
    }
  if (c < COLS/8)
    {
    x_coords[total_points]=c;
    y_coords[total_points]=r;
    total_points++;
    }
  for (c=COLS-4; c>=COLS*7/8; c--)
    {
    RedGrad=Raw[(r*COLS+c)*3+0]-Raw[(r*COLS+c+3)*3+0];
    GreenGrad=Raw[(r*COLS+c)*3+1]-Raw[(r*COLS+c+3)*3+1];
    BlueGrad=Raw[(r*COLS+c)*3+2]-Raw[(r*COLS+c+3)*3+2];
    if (RedGrad+GreenGrad+BlueGrad > CROSSOVER_THRESHOLD)
      break;
    }
  if (c >= COLS*7/8)
    {
    x_coords[total_points]=c;
    y_coords[total_points]=r;
    total_points++;
    }
  }
if (total_points < 10)
  {
//  MessageBox(NULL,"Unable to find boundary points for mask","MakeMask()",MB_OK | MB_APPLMODAL);
  return;
  }
Fit_Circle_To_Points(total_points,x_coords,y_coords,MaskCircle);
MaskCircle[2]-=2.0;

if (MaskImage != NULL)
  MakeStareMask(MaskTopRow,MaskBottomRow,MaskCircle,ROWS,COLS,MaskImage);

free(top_line_count);
free(bottom_line_count);
}



		/*********************************************************
		** This function creates a binary image (ones and zeroes),
		** where one means the pixel is valid, zero means invalid.
		** It takes as input three clipping functions:
		** two horizontal lines, represented by integer row values,
		** and a circle, represented by its floating-point equation.
		** The pixels inside all the clipping functions, or on the
		** boundary (exactly on the clipping functions), are valid.
		*********************************************************/

void MakeStareMask(int top_row,int bottom_row,	/* top and bottom clipping lines */
				   double circle[3],			/* left and right clipping circle */
				   int ROWS,int COLS,			/* size of image */
				   unsigned char *MaskImage)	/* output:  binary image containing mask */

{
int	r,c;
double	radius;

for (r=0; r<ROWS; r++)
  for (c=0; c<COLS; c++)
    {
    MaskImage[r*COLS+c]=0;	/* start off INVALID */
    if (r < top_row)
      continue;
    if (r > bottom_row)
      continue;
    radius=sqrt(SQR((double)r-circle[1])+SQR((double)c-circle[0]));
    if (radius > circle[2])
      continue;
    MaskImage[r*COLS+c]=255;	/* passed all clips -- VALID pixel */
    }
}

